/////////////////////////////////////////////////////////////////////////
//
// © University of Southampton IT Innovation Centre, 2017
//
// Copyright in this software belongs to University of Southampton
// IT Innovation Centre of Gamma House, Enterprise Road, 
// Chilworth Science Park, Southampton, SO16 7NS, UK.
//
// This software may not be used, sold, licensed, transferred, copied
// or reproduced in whole or in part in any manner or form or in or
// on any media by any person other than in accordance with the terms
// of the Licence Agreement supplied with the software, or otherwise
// without the prior written consent of the copyright owners.
//
// This software is distributed WITHOUT ANY WARRANTY, without even the
// implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
// PURPOSE, except where stated in the Licence Agreement supplied with
// the software.
//
//      Created By :            Simon Crowle
//      Created Date :          19-Dec-2017
//      Created for Project :   FLAME
//
/////////////////////////////////////////////////////////////////////////

package uk.ac.soton.innovation.flame.clmc.monspec.test.base;

import java.io.BufferedReader;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.util.ArrayList;
import java.util.List;
import java.util.Properties;
import java.util.concurrent.TimeUnit;
import org.influxdb.InfluxDB;
import org.influxdb.InfluxDBFactory;
import org.influxdb.dto.Query;
import org.influxdb.dto.QueryResult;
import org.influxdb.dto.QueryResult.Result;
import org.influxdb.dto.QueryResult.Series;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;



/**
 * Helper class for accessing InfluxDB and processing results 
 */
public class InfluxUtility
{
    private static Logger     logger = LoggerFactory.getLogger(InfluxUtility.class );
    private static Properties testConfig;
    private static InfluxDB   influxDB;
    
    
    public static Properties getTestConfig()
    {
        if ( testConfig == null )
        {
            InputStream is = null;
            try
            {
                testConfig = new Properties();
                
                is = ClassLoader.class.getResourceAsStream( "/monSpecTestConfig.properties" );
                
                if ( is != null )
                {
                    testConfig.load( is );
                    logger.info( "Found test configuration OK" );
                }
                else
                    throw new Exception( "Failed to find test configuration" );
            }
            catch ( Exception ex )
            {
                String err = "Could not find test configuration: " + ex.getMessage();
                logger.error( err );
            }
            finally
            {
                try
                {
                    if ( is != null )
                        is.close();
                }
                catch ( Exception ex )
                { logger.warn( "Had problems closing resource input stream" ); }
            }
        }
        
        return testConfig;
    }
    
    public static InfluxDB getInfluxDBConnection() throws Exception
    {
        InfluxDB result = null;
        
        try
        {
            if ( influxDB == null )
            {
                Properties config = getTestConfig();
                logger.info( "Trying to connect to Influx: " + config.getProperty("influxDB_EP") );
        
                influxDB = InfluxDBFactory.connect( config.getProperty("influxDB_EP"), 
                                                    config.getProperty("influxDB_UN"),
                                                    config.getProperty("influxDB_PW") );
                
                result = influxDB;
                logger.info( "Influx connection OK" );
            }
            else
                result = influxDB;
        }
        catch ( Exception ex )
        {
            String err = "Failed to load InfluxDB configuration: " + ex.getMessage();
            logger.error( err, ex );
            
            throw new Exception( err );
        }
        
        return result;
    }
    
    public static List<String> getInputSeries( String label ) throws Exception
    {
        // Safety first
        if ( label == null || label.isEmpty() )
            throw new Exception( "Could not get input series: label invalid" );
        
        return tryGetTestFile( "/inputs/" + label );
    }
    
    public static List<String> getQueries( String label ) throws Exception
    {
        // Safety first
        if ( label == null || label.isEmpty() )
            throw new Exception( "Could not get queries: label invalid" );
        
        return tryGetTestFile( "/queries/" + label );
    }
    
    public static List<String> executeQuery( String query, boolean printOut ) throws Exception
    {
        List<String> result = null;
        
        // Safety first
        if ( query == null || query.isEmpty() )
            throw new Exception( "Could not execute query: invalid input param" );
        
        if ( influxDB == null )
            throw new Exception( "Could not execute query: no InfluxDB connection" );
        
        // Query it
        Query fluxQuery = new Query( query, testConfig.getProperty("influxDB_DB") );
        QueryResult qResult = influxDB.query( fluxQuery, TimeUnit.NANOSECONDS );
        
        // Safety
        if ( qResult == null )
            throw new Exception( "Got null InfluxDB query result for: " + query );
        else
            result = parseFluxQueryResult( qResult );
        
        if ( printOut )
        {
            String rOut = "";
            for ( String line : result )
                rOut += line + "\n";
            
            logger.info( rOut );
        }
        
        return result;
    }
    
    // Private methods ---------------------------------------------------------
    private static List<String> tryGetTestFile( String path ) throws Exception
    {
        List<String> result = new ArrayList<>();
        
        // Safety first
        if ( path == null || path.isEmpty() )
            throw new Exception( "Could not get test file: path invalid" );
        
        InputStream    is = null;
        BufferedReader br = null;
        
        try
        {
            // Load resource as stream
            is = ClassLoader.class.getResourceAsStream( path );
            
            if ( is != null )
            {
                // Read into list
                br = new BufferedReader( new InputStreamReader(is) );
                String nextLine;
                
                while( (nextLine = br.readLine()) != null )
                {
                    if ( !nextLine.startsWith("#") && !nextLine.isEmpty() )
                        result.add( nextLine );
                }
            }
            else
                throw new Exception( "Failed to find test file" );
        }
        catch ( Exception ex )
        {
            String err = "Could not find test file: " + ex.getMessage();
            logger.error( err );
        }
        // Clean up
        finally
        {
            try
            {
                if ( br != null )
                    br.close();
                
                if ( is != null )
                    is.close();
            }
            catch ( Exception ex )
            { logger.warn( "Had problems closing resource input stream: " + ex.getMessage() ); }
        }
        
        // Final safety check
        if ( result.isEmpty() )
            throw new Exception( "Found no lines in test file" );
        
        return result;
    }
    
    /**
     * This method simply parses the result of the query and formats as CSV (tab delimited)
     * 
     * @param queryResult  - The structured query result returned by InfluxDB
     * @return             - A list of rows in Line Protocol format
     */
    private static List<String> parseFluxQueryResult( QueryResult queryResult )
    {
        List<String> result = new ArrayList<>();
        
        for ( Result res : queryResult.getResults() )
        {
            for ( Series series : res.getSeries() )
            {
                // Result series name
                result.add( series.getName() );
                
                // TODO: Unclear what the tags collection is for at the moment
                
                // Column titles
                String colNames = "";
                int volInd = 0;
                for ( String colName : series.getColumns() )
                {
                    // Special formatting for time value
                    if ( volInd == 0 )
                        colNames += colName + "\t\t\t";
                    else
                        colNames += colName + "\t";
                    
                    ++volInd;
                }
                    
                
                colNames = colNames.substring( 0, colNames.length() -1 ); // Trim last comma
                result.add( colNames );
                
                // Results
                for ( List<Object> rowList : series.getValues() )
                {
                    String valList = "";
                    volInd = 0;
                    
                    for ( Object val : rowList )
                    {
                        // Special formatting for time value
                        // Note: some nasty rounding errors in nanosecond number component
                        // for now round down to millisecond
                        if ( volInd == 0 )
                        {
                            Double tD = (Double) val;
                            Double msFrac = tD / 100000;
                            Long tL = msFrac.longValue();
                            tL *= 100000;
                            
                            valList += tL.toString() + "\t";
                            
                        }
                        else
                            valList += val.toString() + "\t";
                        
                        ++volInd;
                    }
                    
                    valList = valList.substring( 0, valList.length() -1 ); // Trim last comma
                    result.add( valList );
                }
            }
        }
        
        return result;
    }
}
